(* motor (main loop) *)
(* 2020-03-09 solyga@gmx.de *)
IMPLEMENTATION MODULE Mot;

IMPORT WinDef, WinNT, WinBase;
IMPORT Lib2, Lib3, Cfg, SO, Log;
FROM SYSTEM IMPORT CAST;

TYPE
  THandle = WinDef.HANDLE;

CONST
  mid = Cfg.midMot;

VAR
  tmr: THandle;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, FALSE );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, FALSE );
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  Lib3.Halt();
END Fatal;


PROCEDURE WaitWinEvent( h: THandle );
VAR
  i: LONGCARD;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  i := WinBase.WaitForSingleObject( h, WinBase.INFINITE );
  IF i = MAX( LONGCARD ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'WaitWinEvent: WaitForSingleObject failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  IF i > 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'WaitWinEvent: WaitForSingleObject returned ' );
    SO.WrLngCard( b, i, 1 );
    Fatal( s );
  END;
END WaitWinEvent;

PROCEDURE NewWinTimer(): THandle;
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateWaitableTimer( NIL, FALSE, NIL );  (* attrib, manual, name *)
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewWinTimer: CreateWaitableTimer failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  RETURN h;
END NewWinTimer;

PROCEDURE StartWinTimer( h: THandle; dt: LONGCARD; cyc: BOOLEAN );  (* ms *)
CONST
  dtMax = 429496;  (* LowPart = 0, HighPart = 1 ==> 429496.7296 ms *)
VAR
  per: LONGINT;
  due: WinNT.LARGE_INTEGER;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF dt > dtMax THEN
    SO.Init( b, s );
    SO.WrStr( b, 'StartWinTimer: Invalid period' );
    Fatal( s );
  END;
  per := LONGINT( dt );
  due.HighPart := -1;  (* relative time *)
  due.LowPart  := CAST( LONGCARD, -1 * per * 1000 * 10 );
  IF ~ cyc THEN per := 0 END;
  IF ~ WinBase.SetWaitableTimer( h, due, per, NIL, NIL, FALSE ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'StartWinTimer: SetWaitableTimer failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END StartWinTimer;

PROCEDURE StopWinTimer( h: THandle );  (* does not clear the signal state *)
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ WinBase.CancelWaitableTimer( h ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'StopWinTimer: CancelWaitableTimer failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END StopWinTimer;

CONST WaitWinTimer = WaitWinEvent;


PROCEDURE Wait();
BEGIN
  WaitWinTimer( tmr );
  Lib2.IncL( cnt, 1 );
END Wait;

PROCEDURE Exec( p: PROC );
BEGIN
  IF ( p = NIL ) OR ( p = NULLPROC ) THEN Fatal( 'Exec: Invalid procedure' ) END;
  LOOP
    Wait();
    p();
  END;
END Exec;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule: ' );
  SO.WrLngCard( b, cnt, 1 );
  Message( s );
  oldTerm();
END DoneModule;

PROCEDURE InitModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  Log.Register( mid, 'Mot' );
  cnt := 0;
  tmr := NewWinTimer();
  StartWinTimer( tmr, period, TRUE );

  SO.Init( b, s );
  SO.WrStr( b, 'InitModule: ' );
  SO.WrLngCard( b, cnt, 1 );
  Message( s );

  Lib3.Terminate( DoneModule, oldTerm );
END InitModule;

BEGIN
  InitModule();
END Mot.
